from time import sleep
import requests
import numpy as np
from .exceptions import DatameshConnectError
import os

DATAMESH_READ_TIMEOUT = os.getenv("DATAMESH_READ_TIMEOUT", 10)
DATAMESH_READ_TIMEOUT = None if DATAMESH_READ_TIMEOUT == "None" else float(DATAMESH_READ_TIMEOUT)
DATAMESH_CONNECT_TIMEOUT = os.getenv("DATAMESH_CONNECT_TIMEOUT", 3.05)
DATAMESH_CONNECT_TIMEOUT = None if DATAMESH_CONNECT_TIMEOUT == "None" else float(DATAMESH_CONNECT_TIMEOUT)
DATAMESH_WRITE_TIMEOUT = os.getenv("DATAMESH_WRITE_TIMEOUT", "None")
DATAMESH_WRITE_TIMEOUT = None if DATAMESH_WRITE_TIMEOUT == "None" else float(DATAMESH_WRITE_TIMEOUT)

def retried_request(url, method="GET", data=None, params=None, headers=None, retries=8, timeout=(DATAMESH_CONNECT_TIMEOUT, DATAMESH_READ_TIMEOUT)):
    """
    Retried request function with exponential backoff

    Parameters
    ----------
    url : str
        URL to request
    method : str, optional
        HTTP method, by default "GET"
    data : str, optional
        Request data, by default None
    headers : dict, optional
        Request headers, by default None
    retries : int, optional
        Number of retries, by default 8
    timeout : tupe(float, float), optional
        Request connect and read timeout in seconds, by default (3.05, 10)

    Returns
    -------
    requests.Response
        Response object

    Raises
    ------
    requests.RequestException
        If request fails

    """
    retried = 0
    while retried < retries:
        try:
            resp = requests.request(
                method=method, url=url, data=data, params=params, headers=headers, timeout=timeout
            )
            # Bad Gateway results in waiting for 10 seconds
            # and retrying
            if resp.status_code == 502:
                sleep(30)
                raise requests.RequestException
        except (
            requests.RequestException,
            requests.ReadTimeout,
            requests.ConnectionError,
            requests.ConnectTimeout,
        ) as e:
            sleep(0.1 * 2**retried)
            retried += 1
            if retried == retries:
                raise DatameshConnectError(f"Failed to connect to {url} after {retries} retries with error: {e}")
        else:
            return resp



"""Console script for oceanum library."""

import click
import os

from ..common.models import ContextObject, TokenResponse, Auth0Config

@click.group()
@click.pass_context
def main(ctx: click.Context):
    domain = os.getenv('OCEANUM_DOMAIN', 'oceanum.io')
    if domain == 'oceanum.io':
        auth0_config = Auth0Config(
            domain='auth.oceanum.io',
            client_id='pzXujqmFdkAaVrnsHjn1R7N55GzoIDV2'
        )
    else:
        auth0_config = Auth0Config(
            domain='oceanum-test.au.auth0.com',
            client_id='LTUnEbjjS8Pn1ZTj7VnyYkcTXRShizO9'
        )
    ctx.obj = ContextObject(
        domain=domain, 
        token=TokenResponse.load(domain=domain), 
        auth0=auth0_config
    )
    click.help_option('-h', '--help')




from pathlib import Path
from datetime import datetime, timezone
from platformdirs import user_data_dir

USER_DATA_DIR = Path(user_data_dir('oceanum',  'Oceanum LTD.'))

def bytes_to_human(size):
    """Convert bytes to human-readable format."""
    if size < 1024:
        return f"{size} B"
    elif size < 1024**2:
        return f"{size / 1024:.1f} KiB"
    elif size < 1024**3:
        return f"{size / 1024 ** 2:.1f} MiB"
    elif size < 1024**4:
        return f"{size / 1024 ** 3:.1f} GiB"
    else:
        return f"{size / 1024 ** 4:.1f} TiB"


def item_to_long(item, human_readable=False):
    """Convert item to long listing format."""
    size = item["size"]
    if human_readable:
        size = bytes_to_human(size)
    elif size == 0:
        size = ""
    modified = item["modified"] or ""
    return f"{size:>10}  {modified:>32}  {item['name']}", item["size"]
import time
time.tzname


def format_dt(dt:datetime|str, fmt:str=r'%x %X %Z') -> str:
    """
    Localize datetime to the machine timezone and return a string with language localized date format.

    Args:
        dt (datetime|str): Datetime object or ISO format string.
        fmt (str): Output datetime.strftime pattern.

    Returns:
        str: Localized datetime string.
    
    """
    if isinstance(dt, str):
        dt = datetime.fromisoformat(dt.replace('Z', '+00:00'))
    
    return dt.astimezone().strftime(fmt)



import click

from ...storage import filesystem

from ..main import main
from ..common.utils import bytes_to_human, item_to_long
from ..auth import login_required

# =====================================================================================
# Storage CLI
# =====================================================================================
@main.group()
def storage():
    """Oceanum Storage commands"""
    click.help_option('-h', '--help')

@storage.command()
@click.option("-l", "--long", is_flag=True, help="Long listing format")
@click.option("-h", "--human-readable", is_flag=True, help="Readable sizes with -l")
@click.option("-r", "--recursive", is_flag=True, help="List subdirectories recursively")
@click.argument("path", default="/")
@click.pass_context
@login_required
def ls(
    ctx: click.Context,
    path: str,
    long: bool,
    human_readable: bool,
    recursive: bool,
):
    """List contents in the oceanum storage (the root directory by default)."""
    contents = filesystem.ls(
        path=path, 
        recursive=recursive, 
        detail=long, 
        token=f"Bearer {ctx.obj.token.access_token}", 
        service=f"https://storage.{ctx.obj.domain}/",
    )
    if long:
        sizes = 0
        for item in contents.values(): # type: ignore
            line, size = item_to_long(item, human_readable=human_readable)
            sizes += size
            click.echo(line)
        click.echo(
            f"TOTAL: {len(contents)} objects, {sizes} bytes ({bytes_to_human(sizes)})"
        )
    else:
        click.echo("\n".join(contents))


@storage.command()
@click.option("-r", "--recursive", is_flag=True, help="Copy directories recursively")
@click.argument("source")
@click.argument("dest")
@click.pass_context
def get(
    ctx: click.Context,
    recursive: bool,
    source: str,
    dest: str,
):
    """Copy content from SOURCE to DEST."""
    filesystem.get(
        source=source,
        dest=dest,
        recursive=recursive,
        token=f"Bearer {ctx.obj.token.access_token}", 
        service=f"https://storage.{ctx.obj.domain}/",
    )


@storage.command()
@click.option("-r", "--recursive", is_flag=True, help="Copy directories recursively")
@click.argument("source")
@click.argument("dest")
@click.pass_context
#
def put(
    ctx: click.Context,
    recursive: bool,
    source: str,
    dest: str,
):
    """Copy content from SOURCE to DEST."""
    filesystem.put(
        source=source,
        dest=dest,
        recursive=recursive,
        token=f"Bearer {ctx.obj.token.access_token}", 
        service=f"https://storage.{ctx.obj.domain}/",
    )


@storage.command()
@click.option("-r", "--recursive", is_flag=True, help="Remove directories recursively")
@click.argument("path")
@click.pass_context
def rm(ctx: click.Context, recursive: bool, path: str):
    """Remove PATH."""
    filesystem.rm(
        path=path,
        recursive=recursive,
        token=f"Bearer {ctx.obj.token.access_token}", 
        service=f"https://storage.{ctx.obj.domain}/",
    )


from pydantic import BaseModel
from typing import Optional
from datetime import datetime, timedelta
from .exceptions import DatameshConnectError, DatameshSessionError
from .utils import retried_request
import atexit
import os


class Session(BaseModel):
    id: str
    user: str
    creation_time: datetime
    end_time: datetime
    write: bool
    verified: bool = False

    @classmethod
    def acquire(cls,
                connection):
        """
        Acquire a session from the connection.

        Parameters
        ----------
        connection : Connection
            Connection object to acquire session from.
        """

        # Back-compatibility with beta version (returning dummy session object)
        if not connection._is_v1:
            session =\
                cls(id="dummy_session",
                    user="dummy_user",
                    creation_time=datetime.now(),
                    end_time=datetime.now()+timedelta(hours=connection._session_params.get("duration", 1)),
                    write=False,
                    verified=False)
            session._connection = connection
            atexit.register(session.close)
            return session
        # v1
        try:
            headers = connection._auth_headers.copy()
            headers["Cache-Control"] = "no-store"
            res = retried_request(f"{connection._gateway}/session/",
                                  params=connection._session_params,
                                  headers=headers)
            if res.status_code != 200:
                raise DatameshConnectError("Failed to create session with error: " + res.text)
            session = cls(**res.json())
            session._connection = connection
            atexit.register(session.close)
            return session
        except Exception as e:
            raise DatameshSessionError(f"Error when acquiring datamesh session {e}")
        
    @classmethod
    def from_proxy(cls,
                   session_duration: Optional[float]=None):
        """
        Convenience constructor to acquire a session directly from the proxy.
        Uses environment variables only and used for internal purposes.
        """

        try:
            res = retried_request(f"{os.environ['DATAMESH_ZARR_PROXY']}/session/",
                                  params={'duration': session_duration or 1},
                                  headers={"X-DATAMESH-TOKEN": os.environ['DATAMESH_TOKEN'],
                                           "USER": os.environ['DATAMESH_USER'],
                                           'Cache-Control': 'no-cache'})
            if res.status_code != 200:
                raise DatameshConnectError("Failed to create session with error: " + res.text)
            session = cls(**res.json())
            session._connection = lambda: None
            session._connection._gateway = os.environ['DATAMESH_ZARR_PROXY']
            session._connection._is_v1 = True
            atexit.register(session.close)
            return session
        except Exception as e:
            raise DatameshSessionError(f"Error when acquiring datamesh session from proxy {e}")

    @property
    def header(self):
        return {"X-DATAMESH-SESSIONID": self.id}

    def add_header(self, headers: dict):
        headers.update(self.header)
        return headers
    
    def close(self, finalise_write: bool = False):
        # Back-compatibility with beta version (ignoring)
        if not self._connection._is_v1:
            return
        # datamesh v1
        try:
            atexit.unregister(self.close)
        except:
            pass
        res = retried_request(f"{self._connection._gateway}/session/{self.id}",
                              method="DELETE",
                              params={"finalise_write": finalise_write},
                              headers=self.header)
        if res.status_code != 204:
            if finalise_write:
                raise DatameshConnectError("Failed to finalise write with error: " + res.text)
            print("Failed to close session with error: " + res.text)
    
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # When using context manager, close the session
        # and finalise the write if no exception was raised
        self.close(finalise_write=exc_type is None)



class DatameshConnectError(Exception):
    oceanum_exc = True


class DatameshQueryError(Exception):
    oceanum_exc = True


class DatameshWriteError(Exception):
    oceanum_exc = True


class DatameshSessionError(Exception):
    oceanum_exc = True



import sys

import click

# When adding a new module, must be imported here

from importlib.metadata import entry_points

from oceanum.cli.main import main
from oceanum.cli import auth, datamesh, storage

# for cli_ep in entry_points(group='oceanum.cli'):
#     cli_ep.load()

for run_ep in entry_points(group='oceanum.cli.prax'):
    try:
        run_ep.load()
    except ModuleNotFoundError as e:
        pass



import base64
import time
from typing import Literal
from pathlib import Path
from datetime import datetime

from typing import ClassVar
from typing_extensions import Self
import json

from pydantic import BaseModel, Field

from . import utils


class DeviceCodeResponse(BaseModel):
    device_code: str
    user_code: str
    verification_uri: str
    expires_in: int
    interval: int
    verification_uri_complete: str

class TokenResponse(BaseModel):
    access_token: str
    id_token: str|None = None
    refresh_token: str|None = None
    scope: str|None = None
    domain: str = 'oceanum.tech'
    expires_in: int
    token_type: str
    created_at: datetime = Field(default_factory=datetime.now)

    @classmethod
    def _get_path(cls, domain: str) -> Path:
        domain_slug = domain.replace('.','-')
        return utils.USER_DATA_DIR / f'token-{domain_slug}.json'
    
    @classmethod
    def load(cls, domain: str) -> Self|None:
        token_path = cls._get_path(domain)
        if token_path.exists():
            with token_path.open() as f:
                return cls(**json.load(f))
    
    @property
    def path(self) -> Path:
        return self._get_path(self.domain)

    @property
    def active_org(self) -> str|None:
        if self.access_token is not None:
            payload = base64.b64decode(self.access_token.split('.')[1]+'==').decode('utf-8')
            payload_dict = json.loads(payload)
            return payload_dict.get('https://oceanum.io/active_org', None)
        
    @property
    def email(self) -> str|None:
        if self.access_token is not None:
            payload = base64.b64decode(self.access_token.split('.')[1]+'==').decode('utf-8')
            payload_dict = json.loads(payload)
            return payload_dict.get('https://oceanum.io/email', None)
        
    @property
    def is_expired(self) -> bool:
        return self.created_at.timestamp() + self.expires_in < time.time()
                
    def save(self) -> None:
        if not self.path.parent.exists():
            self.path.parent.mkdir(parents=True, exist_ok=True)
        with self.path.open('w') as f:
            f.write(self.model_dump_json())

    def delete(self) -> bool:
        if self.path.exists():
            self.path.unlink()
            return True
        return False

class Auth0Config(BaseModel):
    domain: str
    client_id: str

class ContextObject(BaseModel):
    domain: str
    token: TokenResponse|None=None
    auth0: Auth0Config|None=None
    